@@ -1,11 +1,13 @@
 package WWW::LinkChecker::Internal::API::Worker;
-$WWW::LinkChecker::Internal::API::Worker::VERSION = '0.12.0';
+$WWW::LinkChecker::Internal::API::Worker::VERSION = '0.14.0';
 use strict;
 use warnings;
 use 5.014;
 
 use Moo;
 
+use Heap::Elem::Str qw( StrElem );
+use Heap::Fibonacci ();
 use JSON::MaybeXS   qw( decode_json encode_json );
 use List::Util 1.34 qw/ any none /;
 
@@ -43,20 +45,24 @@ sub run
         +( $state_fn && ( -e $state_fn ) )
         ? decode_json( path($state_fn)->slurp_utf8 )
         : {
-        stack            => [ { url => $start_url, from => undef(), } ],
-        encountered_urls => { $start_url => 1, },
+        stack            => scalar( Heap::Fibonacci->new() ),
+        encountered_urls => { $start_url => undef(), },
         };
-    my $stack            = $state->{stack};
+    my $stack = $state->{stack};
+    {
+        my $el = StrElem($start_url);
+        $stack->add($el);
+    }
     my $encountered_urls = $state->{encountered_urls};
     my $prev;
     my $dest_url;
     my $url;
 STACK:
 
-    while ( my $url_rec = pop( @{$stack} ) )
+    while ( defined( my $url_rec = $stack->extract_top() ) )
     {
         $dest_url = undef;
-        $url      = $url_rec->{'url'};
+        $url      = $url_rec->val();
         $check_url_inform_cb->( { url => $url, } );
 
         my $mech = WWW::Mechanize->new();
@@ -64,12 +70,12 @@ STACK:
 
         if ($@)
         {
-            push @{$stack}, $url_rec;
+            $stack->add($url_rec);
             if ($state_fn)
             {
                 path($state_fn)->spew_utf8( encode_json($state) );
             }
-            my $from = ( $url_rec->{from} // "START" );
+            my $from = ( $encountered_urls->{$dest_url} // "START" );
             die "SRC URL $from points to '$url'.";
         }
 
@@ -85,8 +91,9 @@ STACK:
                 and $dest_url =~ m{\A\Q$base_url\E}ms
                 and ( none { $dest_url =~ $_ } @before_insert_skips_regexes ) )
             {
-                $encountered_urls->{$dest_url} = 1;
-                push @{$stack}, { url => $dest_url, from => $url, };
+                $encountered_urls->{$dest_url} = $url;
+                my $el = StrElem($dest_url);
+                $stack->add($el);
             }
         };
         foreach my $link ( $mech->links() )
@@ -151,7 +158,7 @@ WWW::LinkChecker::Internal::API::Worker - API object
 
 =head1 VERSION
 
-version 0.12.0
+version 0.14.0
 
 =head1 SYNOPSIS
 
@@ -281,7 +288,7 @@ feature.
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is Copyright (c) 2023 by Shlomi Fish.
+This software is Copyright (c) 2024 by Shlomi Fish.
 
 This is free software, licensed under:
 
